using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;

namespace JSLintForResharper.JSLint
{
	public class ScriptRunner : IDisposable
	{
#if DEBUG
	    private const int Timeout = 60000;
#else
        private const int Timeout = 5000;
#endif
        private StringBuilder errorOuput = new StringBuilder();
		private Process process;
		private AutoResetEvent errorWaitHandle = new AutoResetEvent(false);
	    private bool disposed = false;

		public static LintWarningList Evaluate(string input, string options)
		{
		    var output = GetOutput(input, options);
		    return LintWarningList.ParseAll(output ?? "");
		}

	    public static string GetOutput(string input, string options)
	    {
	        string output;
	        using (var runner = new ScriptRunner())
	        {
	            output = runner.Run(input, options);
	        }
	        return output;
	    }

	    private ScriptRunner()
		{
		}

		private string Run(string input, string options)
		{
			Start(options);
			WriteInput(input);
			BeginReading();
			WaitForExit();
			return errorOuput.ToString();
		}

		private void Start(string optionsString)
		{
			var pi = CreateProcessStart(optionsString);
			process = Process.Start(pi);
			process.ErrorDataReceived += AppendErrorData;
		}

		private void WriteInput(string input)
		{
			process.StandardInput.Write(input);
			process.StandardInput.Flush();
			process.StandardInput.Close();
		}

		private void BeginReading()
		{
			process.BeginErrorReadLine();
		}

		private void AppendErrorData(object sender, DataReceivedEventArgs e)
		{
		    if (disposed) return;
			if (e.Data == null)
				errorWaitHandle.Set();
			else
				errorOuput.AppendLine(e.Data);
		}

		private void WaitForExit()
		{
			errorWaitHandle.WaitOne(Timeout);
			process.WaitForExit(Timeout);
		}

		private static ProcessStartInfo CreateProcessStart(string optionsString)
		{
			return new ProcessStartInfo("cscript.exe", "/nologo \"" + ScriptStorage.LintPath + "\" " + optionsString)
			{
			    CreateNoWindow = true,
			    UseShellExecute = false,
			    RedirectStandardOutput = false,
			    RedirectStandardError = true,
			    RedirectStandardInput = true,
				ErrorDialog = false
			};
		}

		public void Dispose()
		{
			if (process != null)
				process.Dispose();
			if (errorWaitHandle != null)
				((IDisposable)errorWaitHandle).Dispose();
		    disposed = true;
		}
	}
}