﻿using System;
using System.Collections.Generic;
using System.Linq;
using JSLintForResharper.Infrastructure;
using JSLintForResharper.JSLint;
using JSLintForResharper.Resharper.Highlighting;
using JSLintForResharper.Utils;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.JavaScript.LanguageImpl;
using JetBrains.ReSharper.Psi.JavaScript.Tree;
using JetBrains.ReSharper.Psi.Tree;

namespace JSLintForResharper.Resharper.Deamon
{
	internal class DaemonStageProcess : IDaemonStageProcess
	{
		private readonly IDaemonProcess daemonProcess;
		private readonly List<HighlightingInfo> highlightings = new List<HighlightingInfo>();
		private IJavaScriptFile mainFile;
		private IJavaScriptTreeNode mainNode;
		private List<HighlightedWarning> warnings;
		private string code;
		private CoordinateToOffsetTranslator offsetTranslator;
		private int tabSize = 4;
		private bool isMixed;

		public DaemonStageProcess(IDaemonProcess daemonProcess)
		{
			this.daemonProcess = daemonProcess;
		}

		public IDaemonProcess DaemonProcess
		{
			get { return daemonProcess; }
		}

		public void Execute(Action<DaemonStageResult> commiter)
		{
			if (!daemonProcess.FullRehighlightingRequired)
				return;

			var sourceFile = daemonProcess.SourceFile;
			mainFile = sourceFile.GetPsiFile(JavaScriptLanguage.Instance) as IJavaScriptFile;

			highlightings.Clear();

			if (mainFile != null)
			{
				if (sourceFile.PrimaryPsiLanguage is JavaScriptLanguage)
				{
					isMixed = false;
					mainNode = mainFile;
					LintAndAddHighlightings();
				}
				else
				{
					isMixed = true;
					foreach (var subFile in mainFile.Sections)
					{
						mainNode = subFile;
						LintAndAddHighlightings();
					}
				}
			} 
			
			commiter(new DaemonStageResult(highlightings));
		}

		private IEnumerable<LintWarning> Lint(string input)
		{
			string whiteOption = isMixed ? ",'white':true" : "";
			string options = "{'indent':'" + tabSize.ToString() + "'" + whiteOption + "}";
			var lintWarnings = ScriptRunner.Evaluate(input, options);
			System.Diagnostics.Debug.WriteLine(lintWarnings.ToString());
			return lintWarnings;
		}

		private void LintAndAddHighlightings()
		{
			// TODO: find tabsize

			try
			{
				code = mainNode.GetText();
				offsetTranslator = new CoordinateToOffsetTranslator(code, tabSize);
				warnings = Lint(code).Select(CreateHighlightedWarning).ToList();

				foreach (var highlightedWarning in warnings)
				{
					var nodes = mainNode.FindNodesAt(new TreeOffset(highlightedWarning.Offset)).ToArray();

					if (IsWhitespaceWarning(highlightedWarning))
						AddWhitespaceWarning(highlightedWarning, nodes);
					else
						TryAddWarningOnEvidence(highlightedWarning, nodes);

					if (IsNotHightlighted(highlightedWarning, nodes))
						AddNonWhitespaceWarning(highlightedWarning, nodes);
				}
			}
			catch (Exception ex)
			{
				Log(ex);
			}
		}

		private void AddNonWhitespaceWarning(HighlightedWarning highlightedWarning, ITreeNode[] nodes)
		{
			var node = nodes.FirstOrDefault(n => !n.IsWhitespaceToken());
			AddHighlighting(highlightedWarning, node);
		}

		private void TryAddWarningOnEvidence(HighlightedWarning highlightedWarning, ITreeNode[] nodes)
		{
			for (var i = 0; i < nodes.Length; i++)
			{
				var node = nodes[i];
				ITreeNode parent = null;
				var evidence = highlightedWarning.Warning.Evidence;
				var nodeText = node.GetText();
				var noMatch = nodeText != evidence;
				if (noMatch && evidence.StartsWith(nodeText))
					parent = node.GetContainingNode<ITreeNode>(n => n.GetText() == evidence);
				if (parent != null)
				{
					node = parent;
					nodeText = node.GetText();
				}
				else
					continue;

				AddHighlighting(highlightedWarning, node);
				break;
			}
		}

		private void AddWhitespaceWarning(HighlightedWarning highlightedWarning, ITreeNode[] nodes)
		{
			var node = nodes.FirstOrDefault(n => n.IsWhitespaceToken());
			if (node != null)
				AddHighlighting(highlightedWarning, node);
		}

		private static bool IsWhitespaceWarning(HighlightedWarning highlightedWarning)
		{
			return highlightedWarning.Warning.Message.ToLower().Contains("space");
		}

		private static bool IsNotHightlighted(HighlightedWarning highlightedWarning, ITreeNode[] nodes)
		{
			return !highlightedWarning.Highlighted && nodes.Any();
		}

		private HighlightedWarning CreateHighlightedWarning(LintWarning warning)
		{
			int offset = GetMessageOffset(warning);
			return new HighlightedWarning(warning, offset);
		}

		private void AddHighlighting(HighlightedWarning warning, ITreeNode node)
		{
			var sameHighlight = highlightings.SingleOrDefault(SameNode(node));
			if (sameHighlight != null)
			{
				((LintHighlighting)sameHighlight.Highlighting).AddWarning(warning.Warning);
			}
			else
			{
				var highlightingInfo = new HighlightingInfo(node.GetDocumentRange(), new LintHighlighting(node, warning.Warning));
				highlightings.Add(highlightingInfo);
			}
			warning.Highlighted = true;
		}

		private static Func<HighlightingInfo, bool> SameNode(ITreeNode node)
		{
			return h => ((LintHighlighting)h.Highlighting).Expression == node;
		}

		private int GetMessageOffset(LintWarning message)
		{
			return offsetTranslator.GetOffset(message.Line, message.Character);
		}

		private void Log(Exception ex)
		{
			Log("", ex);
		}

		private void Log(string extra, Exception ex, params object[] args)
		{
			LogProvider.Instance.Log(String.Format("{0}: {1}, {2}" + extra,
				new object[]{
					DateTime.Now.ToString("yy-MM-dd HH:mm:ss"),
					ex.Message + Environment.NewLine + ex.StackTrace,
					daemonProcess.SourceFile.Name
				}.Union(args).ToArray()
			));
		}
	}

	public class HighlightedWarning
	{
		public LintWarning Warning { get; private set; }
		public int Offset { get; private set; }
		public bool Highlighted { get; set; }

		public HighlightedWarning(LintWarning warning, int offset)
		{
			Warning = warning;
			Offset = offset;
		}
	}
}
